// SPDX-License-Identifier: GPL-2.0
/*
* Copyright (C) 2020 Red Hat. All rights reserved.
*
* This file is released under the GPL.
*/

#include "nvfs.h"

static void *nvfs_alloc_new_directory_block(struct inode *dir)
{
	struct nvfs_superblock *nvs;
	uint64_t blk;
	int r;

	nvs = dir->i_sb->s_fs_info;

	r = dquot_alloc_space_nodirty(dir, 1U << nvs->log2_page_size);
	if (unlikely(r))
		return ERR_PTR(r);

	blk = nvfs_new_block(nvs, NULL);
	if (unlikely(!blk)) {
		dquot_free_space_nodirty(dir, 1U << nvs->log2_page_size);
		return ERR_PTR(-ENOSPC);
	}

	nvfs_set_inode_size(i_to_nmi(dir), dir->i_size + (1U << nvs->log2_page_size));

	return nvs->memory_map + blk;
}

static void nvfs_free_directory_block(struct inode *dir, unsigned long block)
{
	struct nvfs_superblock *nvs = dir->i_sb->s_fs_info;
	struct nvfs_dir_head *ndh = nvs->memory_map + block;

	dquot_free_space_nodirty(dir, 1UL << nvs->log2_page_size);

	nv_store(nvs, ndh->magic, cpu_to_le64(0));
	nv_wmb(nvs);
	nvfs_free_blocks(nvs, block, 1);

	nvfs_set_inode_size(i_to_nmi(dir), dir->i_size - (1U << nvs->log2_page_size));
}

static void init_dir_page(struct nvfs_superblock *nvs, struct nvfs_dir_head *ndh, uint64_t ino)
{
	nv_store(nvs, ndh->self, cpu_to_le64((void *)ndh - nvs->memory_map));
	nv_store(nvs, ndh->next, cpu_to_le64(0));
	nv_store(nvs, ndh->inode, cpu_to_le64(ino));
	nv_store(nvs, ndh->nde->name_len_and_type, cpu_to_le32(nvfs_dir_entry_make_size((1UL << nvs->log2_page_size) - NVFS_DIR_OFFSET) << NAME_LEN_OFFSET));
}

static void finish_dir_page(struct nvfs_superblock *nvs, struct nvfs_dir_head *ndh)
{
	nv_wmb(nvs);
	nv_store(nvs, ndh->magic, NVFS_DIR_HEAD_MAGIC);
	nv_memcpy_flushcache(nvs, ndh->dirid, p_superblock(nvs)->dirid, NVFS_DIRID_LEN);
}

int nvfs_init_empty_dir(struct inode *parent, struct inode *dir)
{
	struct nvfs_superblock *nvs;
	struct nvfs_dir_head *ndh;
	struct nvfs_inode *nvi;
	unsigned j;

	ndh = nvfs_alloc_new_directory_block(dir);
	if (unlikely(IS_ERR(ndh)))
		return PTR_ERR(ndh);

	nvs = dir->i_sb->s_fs_info;

	init_dir_page(nvs, ndh, dir->i_ino);
	finish_dir_page(nvs, ndh);

	nvi = i_to_nmi(dir)->nvi;
	for (j = 0; j < NVFS_RADIX_BLOCKS; j++)
		nv_store(nvs, nvi->i_radix[j], cpu_to_le64((void *)ndh - nvs->memory_map));

	nv_store(nvs, nvi->i_parent, cpu_to_le64(parent->i_ino));

	return 0;
}


static struct nvfs_dir_entry *nvfs_prepare_page(struct nvfs_dir_head *ndh, struct dentry *dentry, ino_t ino)
{
	struct nvfs_memory_inode *nmi;
	struct nvfs_superblock *nvs;
	struct nvfs_dir_entry *nde;
	void *end;
	size_t needed_len;

	cond_resched();

	nmi = i_to_nmi(d_inode(dentry->d_parent));
	nvs = nmi->vfs_inode.i_sb->s_fs_info;
	end = (void *)ndh + (1UL << nvs->log2_page_size);
	nde = ndh->nde;
	needed_len = nvfs_dir_entry_needed_size(dentry->d_name.len);
	do {
		size_t de_size = nvfs_dir_entry_size(nde);
		size_t nns;

		if (le32_to_cpu(nde->name_len_and_type) & TYPE_MASK)
			goto used;

		if (nvfs_dir_entry_size(nde) < needed_len) {
			size_t next_size;
			struct nvfs_dir_entry *next_nde = (void *)nde + de_size;
			if ((void *)next_nde >= end)
				goto used;
			if (le32_to_cpu(next_nde->name_len_and_type) & TYPE_MASK)
				goto used;
			next_size = de_size + nvfs_dir_entry_size(next_nde);
			nv_store_cached(nvs, nde->name_len_and_type, cpu_to_le32(nvfs_dir_entry_make_size(next_size) << NAME_LEN_OFFSET));
			nv_flush_cached(nvs, &nde->name_len_and_type, sizeof(nde->name_len_and_type));
			nv_wmb(nvs);
			continue;
		}

		nns = nvfs_dir_entry_needed_size(le32_to_cpu(nde->name_len_and_type));
		if (nns > needed_len) {
			size_t remaining;
			struct nvfs_dir_entry *next_nde;
			remaining = nns - needed_len;
			if (unlikely(remaining < 16))
				goto used;
			next_nde = (void *)nde + needed_len;
			nv_store_cached(nvs, next_nde->name_len_and_type, cpu_to_le32(nvfs_dir_entry_make_size(remaining) << NAME_LEN_OFFSET));
			nv_flush_cached(nvs, &next_nde->name_len_and_type, sizeof(nde->name_len_and_type));
			nv_wmb(nvs);
		}

		if (dentry->d_name.len < 4) {
			__le32 key;
			strncpy((char *)&key, dentry->d_name.name, 4);
			nv_store_cached(nvs, *(__le32 *)nde->name, key);
		} else {
			memcpy(nde->name, dentry->d_name.name, dentry->d_name.len);
		}
		nv_store_cached(nvs, nde->name_len_and_type, cpu_to_le32(dentry->d_name.len << NAME_LEN_OFFSET));
		nv_store_cached(nvs, nde->inode, cpu_to_le64(ino));
		nv_flush_cached(nvs, nde, offsetof(struct nvfs_dir_entry, name[dentry->d_name.len]));
		nv_wmb(nvs);
		return nde;

used:
		nde = (void *)nde + de_size;
	} while ((void *)nde < end);
	if (unlikely((void *)nde > end)) {
		nvfs_error(nvs, "corrupted directory page");
		return ERR_PTR(-EUCLEAN);
	}
	return NULL;
}

static int nvfs_split_page(struct nvfs_superblock *nvs, struct nvfs_dir_head *ndh, struct nvfs_dir_head *ndh1, struct nvfs_dir_head *ndh2, unsigned bit)
{
	struct nvfs_dir_entry *nde, *nde1, *nde2;
	void *end, *end1, *end2;
	size_t left1, left2;

	nde = ndh->nde;
	nde1 = ndh1->nde;
	nde2 = ndh2->nde;
	end = (void *)ndh + (1UL << nvs->log2_page_size);
	do {
		uint32_t hash;
		size_t de_size = nvfs_dir_entry_size(nde);

		if (unlikely(!nvfs_dir_entry_type(nde)))
			goto next_entry;

		hash = nvfs_hash_name(nde->name, nvfs_dir_entry_namelen(nde));
		if (!((hash >> bit) & 1)) {
			memcpy(nde1, nde, de_size);
			nde1 = (void *)nde1 + de_size;
		} else {
			memcpy(nde2, nde, de_size);
			nde2 = (void *)nde2 + de_size;
		}

next_entry:
		nde = (void *)nde + de_size;
	} while ((void *)nde < end);
	if (unlikely((void *)nde > end)) {
		nvfs_error(nvs, "corrupted directory page");
		return -EUCLEAN;
	}

	end1 = (void *)ndh1 + (1UL << nvs->log2_page_size);
	end2 = (void *)ndh2 + (1UL << nvs->log2_page_size);

	left1 = end1 - (void *)nde1;
	left2 = end2 - (void *)nde2;

	if (left1) {
		memmove((void *)ndh1->nde + left1, ndh1->nde, (void *)nde1 - (void *)ndh1->nde);
		nv_store(nvs, ndh1->nde->name_len_and_type, cpu_to_le32(nvfs_dir_entry_make_size(left1)));
	}
	nv_flush(nvs, (void *)ndh1->nde + left1, (void *)nde1 - (void *)ndh1->nde);

	if (left2) {
		memmove((void *)ndh2->nde + left2, ndh2->nde, (void *)nde2 - (void *)ndh2->nde);
		nv_store(nvs, ndh2->nde->name_len_and_type, cpu_to_le32(nvfs_dir_entry_make_size(left2)));
	}
	nv_flush(nvs, (void *)ndh2->nde + left2, (void *)nde2 - (void *)ndh2->nde);

	return 0;
}

static inline __le64 *nvfs_get_directory_run(struct inode *dir)
{
	struct nvfs_memory_inode *nmi = i_to_nmi(dir);
	struct nvfs_inode *nvi = nmi->nvi;
	return nvi->i_radix;
}

static unsigned long hash_to_ptr(__le64 *ptr, unsigned hash, int hash_shift, unsigned bits)
{
	unsigned offs;
	unsigned long block;

	if (hash_shift > 0)
		offs = (hash >> hash_shift);
	else
		offs = (hash << -hash_shift);
	offs &= (1U << bits) - 1;
	block = le64_to_cpu(ptr[offs]);
	return block;
}

#define NVFS_FIND_CHAIN			-1
#define NVFS_FIND_NEXT_LEVEL		-2

struct split_info {
	__le64 *ptr;
	unsigned bits;
	int split_bit;
	unsigned long path[NVFS_MAX_TREE_DEPTH + 1];
};

static struct nvfs_dir_head *nvfs_find_page(struct inode *dir, uint32_t hash, uint32_t *next_hash, struct split_info *si)
{
	struct nvfs_superblock *nvs = dir->i_sb->s_fs_info;
	struct nvfs_dir_head *ndh;
	__le64 *ptr = nvfs_get_directory_run(dir);
	unsigned bits =	NVFS_LOG2_RADIX_BLOCKS;
	int hash_shift = 32 - NVFS_LOG2_RADIX_BLOCKS;
	unsigned long block, block2;
	unsigned path_n = 0;

again:
	block = hash_to_ptr(ptr, hash, hash_shift, bits);

	ndh = nvs->memory_map + block;
	if (likely(ndh->magic == NVFS_DIR_HEAD_MAGIC)) {
		if (unlikely(si != NULL)) {
			int i;
			si->ptr = ptr;
			si->bits = bits;
			si->path[path_n++] = 0;
			si->split_bit = hash_shift > 0 ? NVFS_FIND_NEXT_LEVEL : NVFS_FIND_CHAIN;
			for (i = max(hash_shift, 0); i < hash_shift + bits; i++) {
				uint32_t hash2 = hash ^ (1U << i);
				block2 = hash_to_ptr(ptr, hash2, hash_shift, bits);
				if (block2 == block) {
					si->split_bit = i;
				}
			}
		}
		if (unlikely(next_hash != NULL)) {
			int i;
			for (i = max(hash_shift, 0); i < hash_shift + bits; i++) {
				if (!((hash >> i) & 1)) {
					uint32_t hash2 = hash | (1U << i);
					block2 = hash_to_ptr(ptr, hash2, hash_shift, bits);
					if (block2 == block) {
						hash = hash2;
					} else {
						break;
					}
				}
			}
			if (hash_shift >= 0)
				*next_hash = hash + (1U << hash_shift);
			else
				*next_hash = hash + 1U;
		}
		return ndh;
	}

	if (si)
		si->path[path_n++] = block;

	ptr = (__le64 *)ndh;
	bits = nvs->log2_page_size - 3;
	hash_shift -= bits;

	goto again;
}

struct nvfs_dir_entry *nvfs_prepare_add(struct dentry *dentry, struct inode *inode)
{
	struct nvfs_memory_inode *nmi = i_to_nmi(d_inode(dentry->d_parent));
	struct nvfs_superblock *nvs = nmi->vfs_inode.i_sb->s_fs_info;
	struct nvfs_dir_head *ndh;
	struct nvfs_dir_entry *nde;
	unsigned long n;
	struct split_info si;
	bool need_chain;

	if (unlikely(dentry->d_name.len > nvfs_max_filename_length(nvs)))
		return ERR_PTR(-ENAMETOOLONG);

again:
	need_chain = false;
	ndh = nvfs_find_page(&nmi->vfs_inode, nvfs_hash_name(dentry->d_name.name, dentry->d_name.len), NULL, &si);

	n = 0;
next_page:
	nde = nvfs_prepare_page(ndh, dentry, inode->i_ino);
	if (likely(nde != NULL))
		return nde;

	if (si.split_bit == NVFS_FIND_CHAIN || le64_to_cpu(ndh->next) || need_chain) {
		struct nvfs_dir_head *new_ndh;
		unsigned long page_n;

		need_chain = true;

		page_n = le64_to_cpu(ndh->next);
		if (likely(page_n != 0)) {
			ndh = nvs->memory_map + page_n;
			n++;
			goto next_page;
		}

		if (unlikely(n >= nvfs_max_chain_length(nvs->log2_page_size)))
			return ERR_PTR(-ENOSPC);

		new_ndh = nvfs_alloc_new_directory_block(&nmi->vfs_inode);
		if (unlikely(IS_ERR(new_ndh)))
			return (void *)new_ndh;
		init_dir_page(nvs, new_ndh, nmi->vfs_inode.i_ino);
		finish_dir_page(nvs, new_ndh);
		nv_wmb(nvs);
		nv_store(nvs, ndh->next, cpu_to_le64((void *)new_ndh - nvs->memory_map));
		ndh = new_ndh;
		goto next_page;
	} else if (si.split_bit == NVFS_FIND_NEXT_LEVEL) {
		__le64 *new_page;
		unsigned i;
		uint64_t block = (void *)ndh - nvs->memory_map;

		new_page = nvfs_alloc_new_directory_block(&nmi->vfs_inode);
		if (unlikely(IS_ERR(new_page)))
			return (void *)new_page;

		for (i = 0; i < 1U << (nvs->log2_page_size - 3); i++)
			nv_store(nvs, new_page[i], cpu_to_le64(block));
		nv_wmb(nvs);
		for (i = 0; i < 1U << si.bits; i++) {
			if (le64_to_cpu(si.ptr[i]) == block) {
				nv_store(nvs, si.ptr[i], cpu_to_le64((void *)new_page - nvs->memory_map));
				goto found1;
			}
		}
		nvfs_error(nvs, "corrupted directory tree when creating new level");
		return ERR_PTR(-EUCLEAN);
found1:
		nv_wmb(nvs);
		goto again;
	} else {
		struct nvfs_dir_head *ndh1, *ndh2;
		int r;
		unsigned i, j, k;
		int p;
		uint64_t block;
		struct nvfs_inode *nvi;
		unsigned long copied_path[NVFS_MAX_TREE_DEPTH];

		ndh1 = nvfs_alloc_new_directory_block(&nmi->vfs_inode);
		if (unlikely(IS_ERR(ndh1))) {
			r = PTR_ERR(ndh1);
			goto split_err_0;
		}
		init_dir_page(nvs, ndh1, nmi->vfs_inode.i_ino);

		ndh2 = nvfs_alloc_new_directory_block(&nmi->vfs_inode);
		if (unlikely(IS_ERR(ndh2))) {
			r = PTR_ERR(ndh2);
			goto split_err_1;
		}
		init_dir_page(nvs, ndh2, nmi->vfs_inode.i_ino);

		r = nvfs_split_page(nvs, ndh, ndh1, ndh2, si.split_bit);
		if (unlikely(r)) {
			goto split_err_2;
		}

		finish_dir_page(nvs, ndh1);
		finish_dir_page(nvs, ndh2);

		memset(copied_path, 0, sizeof copied_path);
		for (p = 0; si.path[p]; p++) ;

		block = (void *)ndh - nvs->memory_map;

search_for_block:
		for (i = 0; i < 1U << si.bits; i++) {
			if (le64_to_cpu(si.ptr[i]) == block) {
				goto found2;
			}
		}
		nvfs_error(nvs, "corrupted directory tree when splitting page");
		r = -EUCLEAN;
		goto split_err_3;
found2:
		for (j = i; j < 1U << si.bits; j++) {
			if (le64_to_cpu(si.ptr[j]) != block) {
				break;
			}
		}

		if (p--) {
			__le64 *cp;
			cp = nvfs_alloc_new_directory_block(&nmi->vfs_inode);
			if (unlikely(IS_ERR(cp))) {
				r = PTR_ERR(cp);
				goto split_err_3;
			}
			copied_path[p] = (void *)cp - nvs->memory_map;
			for (k = 0; k < 1UL << (nvs->log2_page_size - 3); k++) {
				unsigned long val;
				if (k >= i && k < (i + j) / 2)
					val = (void *)ndh1 - nvs->memory_map;
				else if (k >= (i + j) / 2 && k < j)
					val = (void *)ndh2 - nvs->memory_map;
				else
					val = le64_to_cpu(si.ptr[k]);
				nv_store(nvs, cp[k], cpu_to_le64(val));
			}
			ndh1 = ndh2 = (void *)cp;
			block = si.path[p];
			if (!p) {
				si.ptr = nvfs_get_directory_run(&nmi->vfs_inode);
				si.bits = NVFS_LOG2_RADIX_BLOCKS;
			} else {
				si.ptr = nvs->memory_map + si.path[p - 1];
				si.bits = nvs->log2_page_size - 3;
			}
			goto search_for_block;
		}

		nvi = nmi->nvi;
		nv_wmb(nvs);
		nvfs_start_tx(nvs);
		for (k = 0; k < NVFS_RADIX_BLOCKS; k++) {
			unsigned long val;
			if (k >= i && k < (i + j) / 2) {
				val = (void *)ndh1 - nvs->memory_map;
				nvfs_add_to_tx(nvs, &nvi->i_radix[k], val, false);
			} else if (k >= (i + j) / 2 && k < j) {
				val = (void *)ndh2 - nvs->memory_map;
				nvfs_add_to_tx(nvs, &nvi->i_radix[k], val, false);
			}
		}
		nvfs_commit_tx(nvs);

		for (p = 0; si.path[p]; p++)
			nvfs_free_directory_block(&nmi->vfs_inode, si.path[p]);
		nvfs_free_directory_block(&nmi->vfs_inode, (void *)ndh - nvs->memory_map);

		goto again;

split_err_3:
		for (p = 0; p < NVFS_MAX_TREE_DEPTH; p++) {
			if (copied_path[p])
				nvfs_free_directory_block(&nmi->vfs_inode, copied_path[p]);
		}
split_err_2:
		nvfs_free_directory_block(&nmi->vfs_inode, (void *)ndh2 - nvs->memory_map);
split_err_1:
		nvfs_free_directory_block(&nmi->vfs_inode, (void *)ndh1 - nvs->memory_map);
split_err_0:
		return ERR_PTR(r);
	}
}

void nvfs_finish_add(struct nvfs_dir_entry *nde, struct dentry *dentry, struct inode *inode)
{
	struct nvfs_memory_inode *nmi = i_to_nmi(inode);
	struct nvfs_superblock *nvs = nmi->vfs_inode.i_sb->s_fs_info;

	nvfs_add_to_tx(nvs, &nde->name_len_and_type, (dentry->d_name.len << NAME_LEN_OFFSET) | (fs_umode_to_ftype(inode->i_mode) + 1) << TYPE_OFFSET, true);
}

int nvfs_add_link(struct dentry *dentry, struct inode *inode)
{
	struct nvfs_dir_entry *nde;

	nvfs_update_time(d_inode(dentry->d_parent), NULL, S_CTIME | S_MTIME);

	nde = nvfs_prepare_add(dentry, inode);

	if (unlikely(IS_ERR(nde)))
		return PTR_ERR(nde);

	nvfs_finish_add(nde, dentry, inode);

	return 0;
}

struct nvfs_dir_entry *nvfs_find_de(struct nvfs_memory_inode *nmi, struct qstr *qname)
{
	struct nvfs_superblock *nvs = nmi->vfs_inode.i_sb->s_fs_info;
	struct nvfs_dir_head *ndh;
	struct nvfs_dir_entry *nde;
	uint64_t next;
	void *end;
	__le32 key;

	strncpy((char *)&key, qname->name, 4);

	ndh = nvfs_find_page(&nmi->vfs_inode, nvfs_hash_name(qname->name, qname->len), NULL, NULL);
next_page:
	end = (void *)ndh + (1UL << nvs->log2_page_size);
	nde = ndh->nde;
	do {
		if (likely(*(__le32 *)nde->name != key))
			goto next_entry;
		if (unlikely(!nvfs_dir_entry_type(nde)))
			goto next_entry;
		if (qname->len != nvfs_dir_entry_namelen(nde))
			goto next_entry;
		if (memcmp(nde->name, qname->name, qname->len))
			goto next_entry;
		return nde;

next_entry:
		nde = (void *)nde + nvfs_dir_entry_size(nde);
	} while ((void *)nde < end);
	if (unlikely((void *)nde > end)) {
		nvfs_error(nvs, "corrupted directory page");
		return ERR_PTR(-EUCLEAN);
	}

	next = le64_to_cpu(ndh->next);
	if (unlikely(next != 0)) {
		ndh = nvs->memory_map + next;
		goto next_page;
	}

	return NULL;
}

struct nvfs_dir_entry *nvfs_prepare_delete(struct nvfs_memory_inode *nmi, struct qstr *qname)
{
	struct nvfs_dir_entry *nde;

	nvfs_update_time(&nmi->vfs_inode, NULL, S_CTIME | S_MTIME);

	nde = nvfs_find_de(nmi, qname);
	if (unlikely(IS_ERR(nde)))
		return nde;

	if (unlikely(!nde)) {
		struct nvfs_superblock *nvs = nmi->vfs_inode.i_sb->s_fs_info;
		nvfs_error(nvs, "delete: directory entry for a file '%s' disappeared", qname->name);
		return ERR_PTR(-EUCLEAN);
	}

	return nde;
}

void nvfs_finish_delete(struct nvfs_superblock *nvs, struct nvfs_dir_entry *nde)
{
	uint32_t name_len_and_type = le32_to_cpu(nde->name_len_and_type);
	name_len_and_type &= NAME_LEN_MASK;

	nv_store_cached(nvs, nde->name_len_and_type, cpu_to_le32(name_len_and_type));
	nv_flush_cached(nvs, &nde->name_len_and_type, sizeof nde->name_len_and_type);
}

void nvfs_finish_delete_tx(struct nvfs_superblock *nvs, struct nvfs_dir_entry *nde)
{
	uint32_t name_len_and_type = le32_to_cpu(nde->name_len_and_type);
	name_len_and_type &= NAME_LEN_MASK;

	nvfs_add_to_tx(nvs, &nde->name_len_and_type, name_len_and_type, true);
}

static struct dentry *nvfs_lookup(struct inode *dir, struct dentry *dentry, unsigned int flags)
{
	struct inode *inode;
	struct nvfs_dir_entry *nde;
	nde = nvfs_find_de(i_to_nmi(dir), &dentry->d_name);
	if (!nde)
		return d_splice_alias(NULL, dentry);
	if (IS_ERR(nde))
		return (void *)nde;
	inode = nvfs_iget(dir->i_sb, le64_to_cpu(nde->inode), NULL);
	if (unlikely(inode == ERR_PTR(-ESTALE))) {
		nvfs_error(dir->i_sb->s_fs_info, "deleted inode referenced");
		return ERR_PTR(-EUCLEAN);
	}
	return d_splice_alias(inode, dentry);
}

static void demangle_pos(struct nvfs_superblock *nvs, loff_t p, uint32_t *hash, unsigned long *page, unsigned *pos)
{
	*hash = p >> 32;
	*page = (uint32_t)p >> (nvs->log2_page_size - 4);
	*pos = p & ((1U << (nvs->log2_page_size - 4)) - 1);
}

static loff_t mangle_pos(struct nvfs_superblock *nvs, uint32_t hash, unsigned long page, unsigned pos)
{
	loff_t p;
	p = pos;
	p |= page << (nvs->log2_page_size - 4);
	p |= (uint64_t)hash << 32;
	return p;
}

static int nvfs_readdir(struct file *file, struct dir_context *ctx)
{
	struct inode *inode = file_inode(file);
	struct nvfs_memory_inode *nmi = i_to_nmi(inode);
	struct nvfs_superblock *nvs = inode->i_sb->s_fs_info;
	struct nvfs_dir_head *ndh;
	uint64_t page_n;
	uint32_t hash, hash_n;
	unsigned long page, this_page;
	unsigned pos;

	if (ctx->pos == -1) {
		goto ret_0;
	}
	if (ctx->pos == 0) {
		if (unlikely(!dir_emit_dot(file, ctx)))
			goto ret_0;
		ctx->pos++;
	}
	if (ctx->pos == 1) {
		if (unlikely(!dir_emit_dotdot(file, ctx)))
			goto ret_0;
		ctx->pos++;
	}

	demangle_pos(nvs, ctx->pos, &hash, &page, &pos);

next_hash:
	ndh = nvfs_find_page(&nmi->vfs_inode, hash, &hash_n, NULL);

	this_page = 0;

next_page:
	if (this_page >= page) {
		struct nvfs_dir_entry *nde;
		void *end;
		end = (void *)ndh + (1UL << nvs->log2_page_size);
		nde = ndh->nde;
		do {
			unsigned type;
			size_t namelen;

			if (((void *)nde - (void *)ndh) >> 4 < pos)
				goto next_entry;

			type = nvfs_dir_entry_type(nde);
			if (unlikely(!type))
				goto next_entry;

			namelen = nvfs_dir_entry_namelen(nde);

			ctx->pos = mangle_pos(nvs, hash, this_page, pos);
			if (!dir_emit(ctx, nde->name, namelen, le64_to_cpu(nde->inode), fs_ftype_to_dtype(type - 1)))
				goto ret_0;

			pos = (((void *)nde - (void *)ndh) >> 4) + 1;

next_entry:
			nde = (void *)nde + nvfs_dir_entry_size(nde);
		} while ((void *)nde < end);
		if (unlikely((void *)nde > end)) {
			nvfs_error(nvs, "corrupted directory page");
			return -EUCLEAN;
		}
		pos = 0;
	}

	page_n = le64_to_cpu(ndh->next);
	if (unlikely(page_n != 0)) {
		ndh = nvs->memory_map + page_n;
		this_page++;
		goto next_page;
	}

	if (hash_n != 0) {
		hash = hash_n;
		page = 0;
		pos = 0;
		goto next_hash;
	}

	ctx->pos = -1;
	return 0;

ret_0:
	return 0;
}

int nvfs_is_dir_empty(struct nvfs_memory_inode *nmi)
{
	struct nvfs_superblock *nvs = nmi->vfs_inode.i_sb->s_fs_info;
	uint32_t hash = 0;
	do {
		struct nvfs_dir_head *ndh;
		struct nvfs_dir_entry *nde;
		uint64_t next;
		void *end;
		ndh = nvfs_find_page(&nmi->vfs_inode, hash, &hash, NULL);
next_page:
		end = (void *)ndh + (1UL << nvs->log2_page_size);
		nde = ndh->nde;
		do {
			if (unlikely(nvfs_dir_entry_type(nde)))
				return 0;
			nde = (void *)nde + nvfs_dir_entry_size(nde);
		} while ((void *)nde < end);
		if (unlikely((void *)nde > end)) {
			nvfs_error(nvs, "corrupted directory page");
			return -EUCLEAN;
		}
		next = le64_to_cpu(ndh->next);
		if (unlikely(next != 0)) {
			ndh = nvs->memory_map + next;
			goto next_page;
		}
	} while (hash != 0);
	return 1;
}

static void nvfs_free_directory_chain(struct inode *dir, unsigned long v)
{
	struct nvfs_superblock *nvs = dir->i_sb->s_fs_info;
	struct nvfs_dir_head *ndh;
	unsigned long next;

next_chain:
	ndh = nvs->memory_map + v;
	if (unlikely(ndh->magic != NVFS_DIR_HEAD_MAGIC)) {
		nvfs_error(nvs, "bad magic on directory");
		return;
	}
	if (unlikely(le64_to_cpu(ndh->inode) != dir->i_ino)) {
		nvfs_error(nvs, "bad inode in directory");
		return;
	}
	next = le64_to_cpu(ndh->next);
	nvfs_free_directory_block(dir, v);
	if (next) {
		v = next;
		goto next_chain;
	}
}

static void nvfs_free_directory_run(struct inode *dir, __le64 *run, unsigned len)
{
	struct nvfs_superblock *nvs = dir->i_sb->s_fs_info;
	unsigned long prev = -1;
	do {
		unsigned long v = le64_to_cpu(*run);
		if (v != prev) {
			struct nvfs_dir_head *ndh = nvs->memory_map + v;
			if (ndh->magic != NVFS_DIR_HEAD_MAGIC) {
				__le64 *next_level = (__le64 *)ndh;
				nvfs_free_directory_run(dir, next_level, 1UL << (nvs->log2_page_size - 3));
				nvfs_free_directory_block(dir, v);
			} else {
				nvfs_free_directory_chain(dir, v);
			}
			prev = v;
		}
	} while (run++, --len);
}

void nvfs_delete_directory(struct inode *dir)
{
	__le64 *run = nvfs_get_directory_run(dir);
	if (likely(le64_to_cpu(*run) != 0))
		nvfs_free_directory_run(dir, run, NVFS_RADIX_BLOCKS);
}

static int nvfs_fsync_directory(struct file *file, loff_t start, loff_t end, int datasync)
{
	struct inode *inode = file_inode(file);
	struct nvfs_superblock *nvs = inode->i_sb->s_fs_info;
	nv_wmb(nvs);
	return 0;
}

const struct inode_operations nvfs_dir_inode_operations = {
	.lookup = nvfs_lookup,
	.create = nvfs_create,
	.mkdir = nvfs_mkdir,
	.symlink = nvfs_symlink,
	.mknod = nvfs_mknod,
	.unlink = nvfs_unlink,
	.rmdir = nvfs_rmdir,
	.rename = nvfs_rename,
	.link = nvfs_link,
	.getattr = nvfs_getattr,
	.setattr = nvfs_setattr,
	.update_time = nvfs_update_time,
	.listxattr = nvfs_listxattr,
	.get_acl = nvfs_get_acl,
	.set_acl = nvfs_set_acl,
};

const struct file_operations nvfs_dir_fops = {
	.llseek = generic_file_llseek,
	.read = generic_read_dir,
	.iterate_shared = nvfs_readdir,
	.fsync = nvfs_fsync_directory,
	.unlocked_ioctl = nvfs_ioctl,
#ifdef CONFIG_COMPAT
	.compat_ioctl = nvfs_compat_ioctl,
#endif
};
