// SPDX-License-Identifier: GPL-2.0
/*
* Copyright (C) 2020 Red Hat. All rights reserved.
*
* This file is released under the GPL.
*/

#include "nvfs.h"

#define FILEID_INO64_GEN64		0x6e
#define FILEID_INO64_GEN64_PARENT	0x6f

static int nvfs_encode_fh(struct inode *inode, __u32 *fh, int *max_len, struct inode *parent)
{
	struct nvfs_inode *nvi, *npi;
	int ml;
	if (!parent)
		ml = 4;
	else
		ml = 8;
	if (*max_len < ml) {
		*max_len = ml;
		return FILEID_INVALID;
	}
	*max_len = ml;
	nvi = i_to_nmi(inode)->nvi;
	fh[0] = inode->i_ino;
	fh[1] = inode->i_ino >> 31 >> 1;
	fh[2] = nvi->i_generation;
	fh[3] = (nvi->i_generation >> 32);
	if (likely(!parent))
		return FILEID_INO64_GEN64;
	npi = i_to_nmi(parent)->nvi;
	fh[4] = parent->i_ino;
	fh[5] = parent->i_ino >> 31 >> 1;
	fh[6] = npi->i_generation;
	fh[7] = (npi->i_generation >> 32);
	return FILEID_INO64_GEN64_PARENT;
}

static struct dentry *nvfs_nfs_get_inode(struct super_block *sb, ino_t ino, uint64_t generation)
{
	struct inode *inode;
	inode = nvfs_iget(sb, ino, &generation);
	if (IS_ERR(inode))
		return ERR_PTR(PTR_ERR(inode));
	return d_obtain_alias(inode);
}

static struct dentry *nvfs_fh_to_dentry(struct super_block *sb, struct fid *fid, int fh_len, int fileid_type)
{
	ino_t ino;
	uint64_t gen;
	if (unlikely(fh_len < 4))
		return NULL;
	if (unlikely(fileid_type != FILEID_INO64_GEN64) && unlikely(fileid_type != FILEID_INO64_GEN64_PARENT))
		return NULL;
	ino = fid->raw[0] + ((ino_t)(fid->raw[1]) << 31 << 1);
	gen = fid->raw[2] + ((uint64_t)fid->raw[3] << 32);
	return nvfs_nfs_get_inode(sb, ino, gen);
}

static struct dentry *nvfs_fh_to_parent(struct super_block *sb, struct fid *fid, int fh_len, int fileid_type)
{
	ino_t ino;
	uint64_t gen;
	if (fh_len < 8)
		return NULL;
	if (unlikely(fileid_type != FILEID_INO64_GEN64_PARENT))
		return NULL;
	ino = fid->raw[4] + ((ino_t)(fid->raw[5]) << 31 << 1);
	gen = fid->raw[6] + ((uint64_t)fid->raw[7] << 32);
	return nvfs_nfs_get_inode(sb, ino, gen);
}

static struct dentry *nvfs_get_parent(struct dentry *child)
{
	struct inode *inode = d_inode(child);
	unsigned long p_ino = le64_to_cpu(i_to_nmi(inode)->nvi->i_parent);
	struct inode *pi = nvfs_iget(child->d_sb, p_ino, NULL);
	return d_obtain_alias(pi);
}

const struct export_operations nvfs_export_operations = {
	.encode_fh		= nvfs_encode_fh,
	.fh_to_dentry		= nvfs_fh_to_dentry,
	.fh_to_parent		= nvfs_fh_to_parent,
	.get_parent		= nvfs_get_parent,
};
