// SPDX-License-Identifier: GPL-2.0
/*
* Copyright (C) 2020 Red Hat. All rights reserved.
*
* This file is released under the GPL.
*/

#include "nvfs.h"

static __always_inline ssize_t nvfs_read_locked(struct nvfs_memory_inode *nmi, struct file *file, char __user *buf, size_t len, bool user, loff_t *ppos)
{
	struct nvfs_superblock *nvs = nmi->vfs_inode.i_sb->s_fs_info;
	ssize_t total = 0;
	loff_t pos;
	int r;
	size_t i_size;

	if (unlikely(!len))
		return 0;

	pos = *ppos;

	i_size = nmi->vfs_inode.i_size;
	if (pos >= i_size)
		return 0;
	if (pos + len > i_size)
		len = i_size - pos;

	if (likely(user) && unlikely(!access_ok(buf, len)))
		return -EFAULT;

	do {
		void *blk, *ptr;

		unsigned page_offset = pos & ((1UL << nvs->log2_page_size) - 1);
		unsigned prealloc = (len + (1UL << nvs->log2_page_size) - 1) >> nvs->log2_page_size;
		unsigned size;

		blk = nvfs_bmap(nmi, pos >> nvs->log2_page_size, &prealloc, NULL, NULL, NULL);
		if (unlikely(IS_ERR(blk))) {
			r = PTR_ERR(blk);
			goto ret_r;
		}
		size = min(len, ((size_t)prealloc << nvs->log2_page_size) - page_offset);
		ptr = blk + page_offset;
		if (unlikely(!blk)) {
			size = min(size, (unsigned)PAGE_SIZE);
			ptr = empty_zero_page;
		}
		if (likely(user))
			size -= __copy_to_user(buf, ptr, size);
		else
			memcpy((__force void *)buf, ptr, size);

		if (unlikely(!size)) {
			r = -EFAULT;
			goto ret_r;
		}

		buf += size;
		len -= size;
		pos += size;
		total += size;
	} while (len);

	r = 0;

ret_r:
	*ppos = pos;

	if (file)
		file_accessed(file);

	return total ? total : r;
}

struct nvfs_aux {
	const char __user *user_buf;
	size_t size;
	unsigned page_offset;
	unsigned copied;
	bool clear;
	bool user;
};

#if defined(CONFIG_X86_64)
static long copy_user_flushcache(void *dst, const void __user *src, unsigned size)
{
	unsigned long flushed, dest = (unsigned long) dst;
	long rc = __copy_user_nocache(dst, src, size, 0);

	/*
	 * __copy_user_nocache() uses non-temporal stores for the bulk
	 * of the transfer, but we need to manually flush if the
	 * transfer is unaligned. A cached memory copy is used when
	 * destination or size is not naturally aligned. That is:
	 *   - Require 8-byte alignment when size is 8 bytes or larger.
	 *   - Require 4-byte alignment when size is 4 bytes.
	 */
	if (size < 8) {
		if (!IS_ALIGNED(dest, 4) || size != 4)
			arch_wb_cache_pmem(dst, size);
	} else {
		if (!IS_ALIGNED(dest, 8)) {
			dest = ALIGN(dest, boot_cpu_data.x86_clflush_size);
			arch_wb_cache_pmem(dst, 1);
		}

		flushed = dest - (unsigned long) dst;
		if (size > flushed && !IS_ALIGNED(size - flushed, 8))
			arch_wb_cache_pmem(dst + size - 1, 1);
	}

	return rc;
}
#endif

static void nvfs_write_page(struct nvfs_superblock *nvs, void *page, unsigned prealloc, void *aux_)
{
	struct nvfs_aux *aux = aux_;
	unsigned total_size = (prealloc << nvs->log2_page_size) - aux->page_offset;
	if (total_size > aux->size)
		total_size = aux->size;

	if (aux->clear)
		nvfs_memzero_flushcache(nvs, page, aux->page_offset);

	if (likely(aux->user)) {
#if defined(CONFIG_X86_64)
		aux->copied = total_size - copy_user_flushcache(page + aux->page_offset, aux->user_buf, total_size);
#else
		/* this is better for multi-threaded workload - 4% improvement */
		aux->copied = total_size - __copy_from_user(page + aux->page_offset, aux->user_buf, total_size);
		arch_wb_cache_pmem(page + aux->page_offset, aux->copied);
#endif
	} else {
		memcpy_flushcache(page + aux->page_offset, (__force void *)aux->user_buf, total_size);
		aux->copied = total_size;
	}

	nv_log(nvs, page + aux->page_offset, aux->copied, __FILE__, __LINE__);

	if (aux->clear)
		nvfs_memzero_flushcache(nvs, page + aux->page_offset + aux->copied, (prealloc << nvs->log2_page_size) - (aux->page_offset + aux->copied));
}

static __always_inline ssize_t nvfs_write_locked(struct nvfs_memory_inode *nmi, struct file *file, const char __user *buf, size_t len, bool user, loff_t *ppos)
{
	struct nvfs_superblock *nvs = nmi->vfs_inode.i_sb->s_fs_info;
	ssize_t total = 0;
	loff_t start_pos, pos;
	int r;
	bool some_new = false;

	if (unlikely(!len))
		return 0;

	if (likely(user) && unlikely(!access_ok(buf, len)))
		return -EFAULT;

	if (file) {
		r = file_modified(file);
		if (unlikely(r < 0))
			return r;
	}

	if (ppos)
		pos = *ppos;
	else
		pos = nmi->vfs_inode.i_size;
	start_pos = pos;

	if (unlikely((uint64_t)pos + len < len) ||
	    unlikely((uint64_t)pos + len > (uint64_t)nvs->s->s_maxbytes)) {
		return -EFBIG;
	}

	while (len) {
		struct nvfs_aux aux;
		void *blk;
		unsigned n_pages;
		bool new;

		aux.user_buf = buf;
		aux.page_offset = pos & ((1UL << nvs->log2_page_size) - 1);
		aux.size = len;
		aux.clear = true;
		aux.user = user;
		n_pages = (aux.page_offset + len + (1UL << nvs->log2_page_size) - 1) >> nvs->log2_page_size;

		blk = nvfs_bmap(nmi, pos >> nvs->log2_page_size, &n_pages, nvfs_write_page, &aux, &new);
		if (unlikely(IS_ERR(blk))) {
			r = PTR_ERR(blk);
			goto ret_r;
		}
		if (unlikely(!blk)) {
			r = -ENOSPC;
			goto ret_r;
		}
		some_new |= new;
		if (!new) {
			aux.clear = false;
			nvfs_write_page(nvs, blk, 1, &aux);
		}

		if (unlikely(!aux.copied)) {
			r = -EFAULT;
			goto ret_r;
		}

		buf += aux.copied;
		len -= aux.copied;
		pos += aux.copied;
		total += aux.copied;
	}

	r = 0;

ret_r:
	if (some_new && unlikely(nmi->was_mmapped)) {
		if (pos > start_pos) {
			uint64_t start_fs_page, end_fs_page, start_page, end_page;

			start_fs_page = start_pos >> nvs->log2_page_size;
			end_fs_page = (pos + (1UL << nvs->log2_page_size) - 1) >> nvs->log2_page_size;

			start_page = start_fs_page << (nvs->log2_page_size - PAGE_SHIFT);
			end_page = end_fs_page << (nvs->log2_page_size - PAGE_SHIFT);

			invalidate_inode_pages2_range(nmi->vfs_inode.i_mapping, start_page, end_page - 1);
		}
	}
	if (pos > nmi->vfs_inode.i_size) {
		nv_wmb(nvs);
		nvfs_set_inode_size(nmi, pos);
	}

	if (ppos)
		*ppos = pos;

	return total ? total : r;
}

static ssize_t read_iter_locked(struct kiocb *iocb, struct iov_iter *to)
{
	struct inode *inode = file_inode(iocb->ki_filp);
	struct nvfs_memory_inode *nmi = i_to_nmi(inode);
	struct nvfs_superblock *nvs = inode->i_sb->s_fs_info;
	ssize_t total = 0;
	loff_t pos = iocb->ki_pos;
	int r;
	int shift = nvs->log2_page_size;
	size_t i_size, count;

	i_size = inode->i_size;
	if (pos >= i_size)
		return 0;
	iov_iter_truncate(to, i_size - pos);

	while ((count = iov_iter_count(to))) {
		void *blk, *ptr;
		size_t page_mask = (1UL << shift) - 1;
		unsigned page_offset = pos & page_mask;
		unsigned prealloc = (count + page_mask) >> shift;
		unsigned size;

		blk = nvfs_bmap(nmi, pos >> shift, &prealloc, NULL, NULL, NULL);
		if (unlikely(IS_ERR(blk))) {
			r = PTR_ERR(blk);
			goto ret_r;
		}
		size = ((size_t)prealloc << shift) - page_offset;
		ptr = blk + page_offset;
		if (unlikely(!blk)) {
			size = min(size, (unsigned)PAGE_SIZE);
			ptr = empty_zero_page;
		}
		size = copy_to_iter(ptr, size, to);
		if (unlikely(!size)) {
			r = -EFAULT;
			goto ret_r;
		}

		pos += size;
		total += size;
	}

	r = 0;

ret_r:
	iocb->ki_pos = pos;

	file_accessed(iocb->ki_filp);

	return total ? total : r;
}

static int iov_iter_map(struct iov_iter *iov, struct iovec *iovec, bool *user)
{
	switch (iov_iter_type(iov)) {
		case ITER_IOVEC:
			*user = true;
			*iovec = iov_iter_iovec(iov);
			return 0;
		case ITER_BVEC:
			*user = false;
			iovec->iov_base = (__force __user void *)kmap(iov->bvec->bv_page) + iov->bvec->bv_offset + iov->iov_offset;
			iovec->iov_len = iov->bvec->bv_len - iov->iov_offset;
			return 0;
		case ITER_KVEC:
			*user = false;
			iovec->iov_base = (__force __user void *)iov->kvec->iov_base + iov->iov_offset;
			iovec->iov_len = iov->kvec->iov_len - iov->iov_offset;
			return 0;
		default:
			WARN_ON_ONCE(1);
			return -EINVAL;
	}
}

static void iov_iter_unmap(struct iov_iter *iov)
{
	if (unlikely(iov_iter_type(iov) == ITER_BVEC))
		kunmap(iov->bvec->bv_page);
}

static ssize_t write_iter_locked(struct kiocb *iocb, struct iov_iter *iov)
{
	ssize_t r = 0;
	ssize_t total = 0;
	struct nvfs_memory_inode *nmi = i_to_nmi(iocb->ki_filp->f_inode);

	while (iov_iter_count(iov)) {
		struct iovec iovec;
		bool user;

		r = iov_iter_map(iov, &iovec, &user);
		if (unlikely(r))
			break;

		r = nvfs_write_locked(nmi, iocb->ki_filp, iovec.iov_base, iovec.iov_len, user, iocb->ki_flags & IOCB_APPEND ? NULL : &iocb->ki_pos);

		iov_iter_unmap(iov);

		if (unlikely(r <= 0))
			break;
		iov_iter_advance(iov, r);

		total += r;
		if (unlikely(r != iovec.iov_len))
			break;
	}

	return total ? total : r;
}

ssize_t nvfs_rw_direct_IO_locked(struct kiocb *iocb, struct iov_iter *iov)
{
	bool wr = iov_iter_rw(iov);
	if (!wr)
		return read_iter_locked(iocb, iov);
	else
		return write_iter_locked(iocb, iov);
}

static ssize_t nvfs_read_iter(struct kiocb *iocb, struct iov_iter *iov)
{
	ssize_t r = 0;
	struct nvfs_memory_inode *nmi = i_to_nmi(iocb->ki_filp->f_inode);

	inode_lock_shared(&nmi->vfs_inode);

	if (!IS_DAX(&nmi->vfs_inode)) {
		r = generic_file_read_iter(iocb, iov);
	} else {
		if (likely(iter_is_iovec(iov)) && likely(!iov->iov_offset) && likely(iov->nr_segs == 1)) {
			/*
			 * This is fast path for the 'read' syscall.
			 * It improves performance by 3%.
			 */
			r = nvfs_read_locked(nmi, iocb->ki_filp, iov->iov->iov_base, iov->count, true, &iocb->ki_pos);
			goto unlock_ret;
		}
		r = read_iter_locked(iocb, iov);
	}

unlock_ret:
	inode_unlock_shared(&nmi->vfs_inode);

	return r;
}

static ssize_t nvfs_write_iter(struct kiocb *iocb, struct iov_iter *iov)
{
	ssize_t r = 0;
	struct nvfs_memory_inode *nmi = i_to_nmi(iocb->ki_filp->f_inode);
	bool shared_lock;
	if (unlikely((iocb->ki_flags & (IOCB_DIRECT | IOCB_APPEND)) == IOCB_DIRECT)) {
		inode_lock_shared(&nmi->vfs_inode);
		if (unlikely(iocb->ki_pos + iov_iter_count(iov) > nmi->vfs_inode.i_size)) {
			inode_unlock_shared(&nmi->vfs_inode);
			goto ex_lock;
		}
		shared_lock = true;
	} else {
ex_lock:
		inode_lock(&nmi->vfs_inode);
		shared_lock = false;
	}

	if (!IS_DAX(&nmi->vfs_inode)) {
		r = __generic_file_write_iter(iocb, iov);
	} else {
		if (likely(iter_is_iovec(iov)) && likely(!iov->iov_offset) && likely(iov->nr_segs == 1)) {
			/*
			 * This is fast path for the 'write' syscall.
			 * It improves performance by 1%.
			 */
			r = nvfs_write_locked(nmi, iocb->ki_filp, iov->iov->iov_base, iov->count, true, iocb->ki_flags & IOCB_APPEND ? NULL : &iocb->ki_pos);
			goto unlock_ret;
		}
		r = write_iter_locked(iocb, iov);
	}

unlock_ret:
	if (unlikely(shared_lock))
		inode_unlock_shared(&nmi->vfs_inode);
	else
		inode_unlock(&nmi->vfs_inode);

	return r;
}

static int nvfs_fsync_file(struct file *file, loff_t start, loff_t end, int datasync)
{
	int r;
	struct inode *inode = file_inode(file);
	struct nvfs_superblock *nvs = inode->i_sb->s_fs_info;
	if (!IS_DAX(inode))
		r = __generic_file_fsync(file, start, end, datasync);
	else
		r = 0;
	nv_wmb(nvs);
	return r;
}

static long nvfs_fallocate(struct file *file, int mode, loff_t offset, loff_t len)
{
	struct inode *inode = file_inode(file);
	int r;

	if (mode & FALLOC_FL_PUNCH_HOLE) {
		inode_lock(inode);
		r = nvfs_punch_hole(i_to_nmi(inode), offset, offset + len);
		inode_unlock(inode);

		return r;
	}

	return -EOPNOTSUPP;
}

#ifdef CONFIG_QUOTA

ssize_t nvfs_quota_read(struct super_block *sb, int type, char *buf, size_t len, loff_t off)
{
	struct nvfs_memory_inode *nmi = i_to_nmi(sb_dqopt(sb)->files[type]);
	ssize_t r;

	r = nvfs_read_locked(nmi, NULL, (__force __user void *)buf, len, false, &off);

	return r;
}

ssize_t nvfs_quota_write(struct super_block *sb, int type, const char *buf, size_t len, loff_t off)
{
	struct nvfs_memory_inode *nmi = i_to_nmi(sb_dqopt(sb)->files[type]);
	ssize_t r;

	r = nvfs_write_locked(nmi, NULL, (__force __user void *)buf, len, false, &off);

	return r;
}

#endif

const struct inode_operations nvfs_file_inode_operations = {
	.fileattr_get = nvfs_fileattr_get,
	.fileattr_set = nvfs_fileattr_set,
	.getattr = nvfs_getattr,
	.setattr = nvfs_setattr,
	.update_time = nvfs_update_time,
	.listxattr = nvfs_listxattr,
	.get_acl = nvfs_get_acl,
	.set_acl = nvfs_set_acl,
};

const struct file_operations nvfs_file_fops = {
	.llseek = generic_file_llseek,
	.read_iter = nvfs_read_iter,
	.write_iter = nvfs_write_iter,
	.splice_read = generic_file_splice_read,
	.splice_write = iter_file_splice_write,
	.fsync = nvfs_fsync_file,
	.fallocate = nvfs_fallocate,
	.mmap = nvfs_mmap,
	.unlocked_ioctl = nvfs_ioctl,
#ifdef CONFIG_COMPAT
	.compat_ioctl = nvfs_compat_ioctl,
#endif
};
