// SPDX-License-Identifier: GPL-2.0
/*
* Copyright (C) 2020 Red Hat. All rights reserved.
*
* This file is released under the GPL.
*/

#include "nvfs.h"

int nvfs_cache_readpage(struct file *file, struct page *page)
{
	struct nvfs_memory_inode *nmi = i_to_nmi(file_inode(file));
	struct nvfs_superblock *nvs = nmi->vfs_inode.i_sb->s_fs_info;
	int r;
	uint64_t block_idx;
	void *data, *pg;

	block_idx = page->index >> (nvs->log2_page_size - PAGE_SHIFT);
	data = nvfs_bmap(nmi, block_idx, NULL, NULL, NULL, NULL);
	if (IS_ERR(data)) {
		r = PTR_ERR(data);
		goto ret;
	}

	pg = kmap(page);
	if (!data) {
		clear_page(pg);
	} else {
		data += (page->index << PAGE_SHIFT) & ((1UL << nvs->log2_page_size) - 1);
		copy_page(pg, data);
		SetPageMappedToDisk(page);
	}
	kunmap(page);

	SetPageUptodate(page);

	r = 0;
ret:
	unlock_page(page);
	return r;
}

int nvfs_cache_writepage(struct page *page, struct writeback_control *wbc)
{
	struct nvfs_memory_inode *nmi = i_to_nmi(page->mapping->host);
	struct nvfs_superblock *nvs = nmi->vfs_inode.i_sb->s_fs_info;
	loff_t i_size = i_size_read(&nmi->vfs_inode);
	pgoff_t end_index = i_size >> PAGE_SHIFT;
	unsigned offset = i_size & (PAGE_SIZE - 1);
	int r;
	uint64_t block_idx;
	void *data, *pg;

	if (page->index >= end_index) {
		if (page->index >= end_index + 1 || !offset) {
			unlock_page(page);
			return 0;
		}
		zero_user_segment(page, offset, PAGE_SIZE);
	}

	block_idx = page->index >> (nvs->log2_page_size - PAGE_SHIFT);
	data = nvfs_bmap(nmi, block_idx, NULL, nvfs_zero_page, NULL, NULL);
	if (IS_ERR(data)) {
		r = PTR_ERR(data);
		mapping_set_error(page->mapping, r);
		goto ret_r;
	}

	pg = kmap(page);
	data += (page->index << PAGE_SHIFT) & ((1UL << nvs->log2_page_size) - 1);
	nv_memcpy_flushcache(nvs, data, pg, PAGE_SIZE);
	kunmap(page);

	r = 0;

ret_r:
	set_page_writeback(page);
	unlock_page(page);
	end_page_writeback(page);

	return r;
}

int nvfs_cache_write_begin(struct file *file, struct address_space *mapping, loff_t pos, unsigned len, unsigned flags, struct page **pagep, void **fsdata)
{
	struct nvfs_memory_inode *nmi = i_to_nmi(file_inode(file));
	struct nvfs_superblock *nvs = nmi->vfs_inode.i_sb->s_fs_info;
	struct page *page;
	int r;
	uint64_t block_idx;
	void *data, *pg;

	page = grab_cache_page_write_begin(mapping, pos >> PAGE_SHIFT, flags);
	if (!page)
		return -ENOMEM;
	*pagep = page;

	if (PageMappedToDisk(page))
		return 0;

	block_idx = page->index >> (nvs->log2_page_size - PAGE_SHIFT);

	data = nvfs_bmap(nmi, block_idx, NULL, nvfs_zero_page, NULL, NULL);
	if (IS_ERR(data)) {
		r = PTR_ERR(data);
		goto err;
	}

	if (!PageUptodate(page)) {
		pg = kmap(page);
		data += (page->index << PAGE_SHIFT) & ((1UL << nvs->log2_page_size) - 1);
		copy_page(pg, data);
		kunmap(page);
		SetPageUptodate(page);
	}

	SetPageMappedToDisk(page);

	return 0;

err:
	unlock_page(page);
	put_page(page);
	*pagep = NULL;
	return r;
}

int nvfs_cache_write_end(struct file *file, struct address_space *mapping, loff_t pos, unsigned len, unsigned copied, struct page *page, void *fsdata)
{
	struct nvfs_memory_inode *nmi = i_to_nmi(file_inode(file));
	/*struct nvfs_superblock *nvs = nmi->vfs_inode.i_sb->s_fs_info;*/
	loff_t size;

	size = pos + copied;
	if (size > nmi->vfs_inode.i_size) {
		nvfs_set_inode_size(nmi, size);
	}

	SetPageUptodate(page);
	set_page_dirty(page);
	unlock_page(page);
	put_page(page);

	return copied;
}

ssize_t nvfs_cached_rw_locked(struct nvfs_memory_inode *nmi, struct file *file, bool wr, char __user *buf, size_t len, loff_t *ppos)
{
	/* new_sync_read */

	struct iovec iov = { .iov_base = buf, .iov_len = len };
	struct kiocb kiocb;
	struct iov_iter iter;
	ssize_t ret;

	init_sync_kiocb(&kiocb, file);
	iov_iter_init(&iter, !wr ? READ : WRITE, &iov, 1, len);
	if (!ppos) {
		kiocb.ki_flags |= IOCB_APPEND;
		kiocb.ki_pos = i_size_read(&nmi->vfs_inode);
	} else {
		kiocb.ki_pos = *ppos;
	}
	//kiocb.ki_flags &= ~IOCB_DIRECT;

	if (!wr)
		ret = generic_file_read_iter(&kiocb, &iter);
	else
		ret = __generic_file_write_iter(&kiocb, &iter);
	if (ppos)
		*ppos = kiocb.ki_pos;
	return ret;
}
