// SPDX-License-Identifier: GPL-2.0
/*
* Copyright (C) 2020 Red Hat. All rights reserved.
*
* This file is released under the GPL.
*/

#include "nvfs.h"

#ifdef CONFIG_NVFS_LOG

#define LOG_BUFFER_SIZE		4194304
#define LOG_BUFFER_RATIO	4

static void nvfs_flush_buffer(struct nvfs_superblock *nvs, size_t l)
{
	ssize_t r;
	if (!l)
		return;
	r = kernel_write(nvs->log_file, nvs->log_buffer, l, NULL);
	if (r < 0)
		printk(KERN_ERR "NVFS: unable to write to log file %s: %d\n", nvs->log_file_name, (int)r);
}

static void nvfs_conditional_flush(struct nvfs_superblock *nvs)
{
	if (nvs->log_buffer && !current->journal_info && nvs->log_buffer_pos >= LOG_BUFFER_SIZE / LOG_BUFFER_RATIO) {
		size_t l;
		mutex_lock(&nvs->log_mutex);
		spin_lock(&nvs->log_spinlock);
		l = nvs->log_buffer_pos;
		spin_unlock(&nvs->log_spinlock);
		if (l)
			nvfs_flush_buffer(nvs, l);
		spin_lock(&nvs->log_spinlock);
		memmove(nvs->log_buffer, nvs->log_buffer + l, nvs->log_buffer_pos - l);
		nvs->log_buffer_pos -= l;
		spin_unlock(&nvs->log_spinlock);
		mutex_unlock(&nvs->log_mutex);
	}
}

static void nvfs_log(struct nvfs_superblock *nvs, void *ptr, size_t len)
{
	if (!nvs->log_buffer)
		return;
	if (len > LOG_BUFFER_SIZE) {
		printk(KERN_ERR "NVFS: too long writes\n");
		return;
	}
	if (nvs->log_buffer_pos + len >= LOG_BUFFER_SIZE) {
		printk(KERN_ERR "NVFS: too many atomic writes\n");
		return;
	}
	memcpy(nvs->log_buffer + nvs->log_buffer_pos, ptr, len);
	nvs->log_buffer_pos += len;
}

void nv_log(struct nvfs_superblock *nvs, void *ptr, size_t len, const char *file, int line)
{
	uint64_t v;
	const char *f;
	f = strrchr(file, '/');
	if (f)
		file = f + 1;
	spin_lock(&nvs->log_spinlock);
	v = ptr - nvs->memory_map;
	//printk("nv_log: %llx, %lx\n", v, len);
	nvfs_log(nvs, &v, sizeof v);
	v = len;
	nvfs_log(nvs, &v, sizeof v);
	strncpy((char *)&v, file, sizeof v);
	nvfs_log(nvs, &v, sizeof v);
	v = line;
	nvfs_log(nvs, &v, sizeof v);
	nvfs_log(nvs, ptr, len);
	spin_unlock(&nvs->log_spinlock);

	nvfs_conditional_flush(nvs);
}

void nv_log_suppress(struct nvfs_superblock *nvs)
{
	BUG_ON(current->journal_info);
	current->journal_info = (void *)1;
}

void nv_log_resume(struct nvfs_superblock *nvs)
{
	BUG_ON(current->journal_info != (void *)1);
	current->journal_info = NULL;
	nvfs_conditional_flush(nvs);
}

__cold int nvfs_init_log(struct nvfs_superblock *nvs)
{
	struct file *f;
	if (!nvs->log_file_name)
		return 0;
	nvs->log_buffer = vmalloc(LOG_BUFFER_SIZE);
	if (!nvs->log_buffer)
		return -ENOMEM;
	f = filp_open(nvs->log_file_name, O_CREAT | O_WRONLY | O_TRUNC | O_APPEND, 0600);
	if (IS_ERR(f)) {
		printk(KERN_ERR "NVFS: unable to open log file %s: %d\n", nvs->log_file_name, (int)PTR_ERR(f));
		return PTR_ERR(f);
	}
	nvs->log_file = f;
	return 0;
}

__cold void nvfs_close_log(struct nvfs_superblock *nvs)
{
	nvfs_flush_buffer(nvs, nvs->log_buffer_pos);
	if (nvs->log_file)
		filp_close(nvs->log_file, NULL);
	vfree(nvs->log_buffer);
}

#endif
